import 'dart:async';

import 'package:flutter/material.dart';

void main() {
  runApp(ImageBuilder());
}

class ImageBuilder extends StatelessWidget {
  ImageBuilder({Key key}) : super(key: key);
  CountDownController countDownController =
      CountDownController(start: Duration(seconds: 30));
  Trigger trigger = Trigger();

  @override
  Widget build(BuildContext context) {
    return MaterialApp(
        home: Container(
            child: Scaffold(
      body: Column(mainAxisAlignment: MainAxisAlignment.center, children: [
        FlatButton(
          child: Text("press to start"),
          onPressed: () {
            trigger.start();
          },
        ),
        Center(
            child: CountDownTimer(
          startInstant: true,
          triggerStart: trigger,
          countDownController: countDownController,
          onStartBuilder: (ctx, duration) {
            return Text("Counter Start Builder");
          },
          onTickBuilder: (ctx, duration) {
            return Text("ticking builder ${duration.inSeconds}");
          },
          onEndBuilder: (ctx, duration) {
            return Text("My counter end Builder");
          },
        ))
      ]),
    )));
  }
}

typedef Widget TriggerStart(BuildContext ctx, Duration duration);
typedef Widget OnEndBuilder(BuildContext ctx, Duration duration);
typedef Widget OnStartBuilder(BuildContext ctx, Duration duration);
typedef Widget OnTickBuilder(BuildContext ctx, Duration duration);
typedef Widget OnErrorBuilder(BuildContext ctx, Duration duration);

class Trigger {
  StreamController triggerStream = new StreamController.broadcast();
  start() {
    triggerStream.add("triggered");
  }

  disposeTrigger() {
    triggerStream.close();
  }
}

class WrongConfiguration extends AssertionError {
  String message;
  WrongConfiguration({this.message}) : super(message);
}

class ControllerBridge {
  StreamController _controller = StreamController.broadcast();
  List<StreamSubscription> subscribers = List();

  link(Function streamReceiver) {
    subscribers.add(_controller.stream.listen((data) {
      streamReceiver(data);
    }));
  }

  add(dynamic data) {
    _controller.add(data);
  }

  close() {
    // subscribers.forEach((s) => s.pause());
    // subscribers.clear();
  }
  dispose() {
    _controller.close();
  }
}

class CountDownController {
  Duration end;
  Duration start;
  Duration tickingStep;
  ControllerBridge controllerBridge = ControllerBridge();
  CountDownController({this.start, this.end, this.tickingStep}) {
    if (start == null) {
      start = Duration(seconds: 60);
    }
    if (tickingStep == null) {
      tickingStep = Duration(seconds: 1);
    }
  }

  void restart() {
    controllerBridge.add("restart");
  }
}

class CountDownTimer extends StatefulWidget {
  CountDownController countDownController;
  OnEndBuilder onEndBuilder;
  OnStartBuilder onStartBuilder;
  OnTickBuilder onTickBuilder;
  Trigger triggerStart;
  bool startInstant;
  OnErrorBuilder onErrorBuilder;
  CountDownTimer(
      {this.onEndBuilder,
      this.onStartBuilder,
      this.onTickBuilder,
      this.countDownController,
      Key key,
      this.triggerStart,
      this.startInstant = true})
      : super(key: key) {
    if (countDownController == null) {
      this.countDownController = CountDownController();
    }
    if (this.triggerStart != null) {
      this.startInstant = false;
    }
    if (this.triggerStart == null && this.startInstant == false) {
      throw new WrongConfiguration(
          message: "trigger cannot be null if instant start is false");
    }
  }

  @override
  _CountDownTimerState createState() => _CountDownTimerState();
}

enum DurationEventType { END, START, ERROR, TICK }

class DurationEvent {
  Duration duration;
  DurationEventType durationEventType;
  DurationEvent({this.duration, this.durationEventType});
}

class _CountDownTimerState extends State<CountDownTimer> {
  StreamController<DurationEvent> streamController =
      new StreamController<DurationEvent>.broadcast();
  Timer timer;
  initState() {
    widget.countDownController.controllerBridge.link((data) {
      if (data == "restart") {
        print("some data");
        partialDispose();
        _build();
      }
    });
    super.initState();
    _build();
  }

  _timerBuild() {
    timer = Timer.periodic(this.widget.countDownController.tickingStep,
        (Timer timer) {
      int dif = this.widget.countDownController.start.inSeconds - timer.tick;
      if (dif <= 0) {
        streamController.sink.add(DurationEvent(
            duration: Duration(seconds: dif),
            durationEventType: DurationEventType.END));
      } else {
        streamController.sink.add(DurationEvent(
            duration: Duration(seconds: dif),
            durationEventType: DurationEventType.TICK));
      }
    });
    streamController.stream.listen((duration) {
      if (duration.durationEventType == DurationEventType.END) {
        timer.cancel();
      }
    });
  }

  _build() {
    if (this.widget.startInstant) {
      print("start auto");
      _timerBuild();
    } else {
      this.widget.triggerStart.triggerStream.stream.listen((data) {
        _timerBuild();
      });
    }
  }

  partialDispose() {
    widget.countDownController.controllerBridge.close();

    timer.cancel();
  }

  void dispose() {
    widget.countDownController.controllerBridge.dispose();
    streamController.close();
    timer.cancel();
    widget.triggerStart.disposeTrigger();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return StreamBuilder<DurationEvent>(
      stream: streamController.stream,
      builder: (ctx, projectSnap) {
        if (projectSnap.connectionState == ConnectionState.none &&
            projectSnap.hasData == null) {
        } else {
          if (projectSnap.hasData) {
            Duration duration = projectSnap.data.duration;
            switch (projectSnap.data.durationEventType) {
              case DurationEventType.END:
                if (this.widget.onEndBuilder != null) {
                  return this.widget.onEndBuilder(context, duration);
                }
                return SizedBox();
                // TODO: Handle this case.
                break;
              case DurationEventType.START:
                if (this.widget.onStartBuilder != null) {
                  return this.widget.onStartBuilder(context, duration);
                }
                return SizedBox();
                // TODO: Handle this case.
                break;
              case DurationEventType.ERROR:
                if (this.widget.onErrorBuilder != null) {
                  return this.widget.onEndBuilder(context, duration);
                }
                return SizedBox();
                // TODO: Handle this case.
                break;
              case DurationEventType.TICK:
                if (this.widget.onTickBuilder != null) {
                  return this.widget.onTickBuilder(context, duration);
                }
                return SizedBox();
                // TODO: Handle this case.
                break;
            }

            return SizedBox();
          }
          if (this.widget.onStartBuilder != null) {
            return this.widget.onStartBuilder(
                context, this.widget.countDownController.tickingStep);
          }
        }
        if (this.widget.onStartBuilder != null) {
          return this.widget.onStartBuilder(
              context, this.widget.countDownController.tickingStep);
        }
        return SizedBox();
      },
    );
  }
}
